package com.cecf.observe;

import com.nrxt.nms.rpc.protobuf.TraceMessage;
import org.apache.flink.api.common.functions.RichMapFunction;
import org.apache.flink.configuration.Configuration;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static com.cecf.observe.OtelUtils.*;

public class PojoSpan2TracesMessage extends RichMapFunction<PojoSpan, byte[]> {
    private Map<String, String> appName2CmdbID;

    @Override
    public void open(Configuration parameters) throws Exception {
        appName2CmdbID = new HashMap<>();
        appName2CmdbID.put("SPRING", "SPRING-CMDB-ID");
        appName2CmdbID.put("OBSERVE", "OBSERVE-CMDB-ID");
        // 电力交易
        appName2CmdbID.put("dljy", "5AF43BBB-FA4F-4392-9564-7C00A4078DA3");
        // 新一代应急指挥系统
        appName2CmdbID.put("sgecsTest", "9FB3802C-8376-48D4-9CF8-EAD8AC398725");
    }

    @Override
    public void close() throws Exception {
    }


    @Override
    public byte[] map(PojoSpan s) throws Exception {
        String cmdbID = appName2CmdbID.get(s.AppName);
        if (cmdbID == null) {
            throw new Exception(String.format("未正确配置的appName:%s没有匹配到cmdbID", s.AppName));
        }
        TraceMessage.APMSpan.SpanTypeCls spanTypeCls;
        switch (s.SpanKind) {
            case SpanKindInternal:
                spanTypeCls = TraceMessage.APMSpan.SpanTypeCls.LOCAL;
                break;
            case SpanKindClient:
            case SpanKindProducer:
                spanTypeCls = TraceMessage.APMSpan.SpanTypeCls.EXIT;
                break;
            case SpanKindServer:
            case SpanKindConsumer:
                spanTypeCls = TraceMessage.APMSpan.SpanTypeCls.ENTRY;
                break;
            default:
                throw new Exception(String.format("未正确配置SpanKind:%d的映射", s.SpanKind));
        }
        List<TraceMessage.Tag> tags = new ArrayList<>();
        for (Map.Entry<String, String> sa : s.SpanAttributes.entrySet()) {
            tags.add(TraceMessage.Tag.newBuilder()
                    .setKey(sa.getKey())
                    .setValue(sa.getValue())
                    .build());
        }
        for (Map.Entry<String, String> sa : s.ResourceAttributes.entrySet()) {
            tags.add(TraceMessage.Tag.newBuilder()
                    .setKey(sa.getKey())
                    .setValue(sa.getValue())
                    .build());
        }


        String peer = s.getAttribute("", String.format("%s:%s", s.getAttribute("", "network.peer.address"),
                s.getAttribute("", "network.peer.peer")));
        String endpoint = s.getAttribute("", "http.route", "messaging.destination.name", "db.operation.name");
        /**
         * 1.JOB，任务触发层；2.METHOD，本地方法层；3.DB，数据库驱动层；4.RPC，RPC框架层；5.HTTP，HTTP框架层；6.MQ，MQ驱动层；7.CACHE，CACHE驱动层
         * */
        TraceMessage.APMSpan.LayerCls layer = TraceMessage.APMSpan.LayerCls.METHOD;
        switch (s.getSpanScopeType()) {
            case SPAN_SCOPE_TYPE_DB:
                layer = TraceMessage.APMSpan.LayerCls.DB;
                break;
            case SPAN_SCOPE_TYPE_HTTP:
                layer = TraceMessage.APMSpan.LayerCls.HTTP;
                break;
            case SPAN_SCOPE_TYPE_MQ:
                layer = TraceMessage.APMSpan.LayerCls.MQ;
                break;
        }

        TraceMessage.APMSpan.Builder entrySpanBuilder = TraceMessage.APMSpan.newBuilder()
                .setTraceId(s.TraceID)
                .setSegmentId("-1")
                .setSpanId(s.SpanID)
                .setSuperSegmentId("-1")
                .setSuperSpanId(s.ParentSpanID == null ? "-1" : s.ParentSpanID)
                .setApmServiceCode(s.ServiceName)
                .setApmServiceInstId(s.getServiceInstanceID())
                .setApmEndpointName(endpoint).
                setCmdbId(cmdbID)
                .setStartTime(s.Timestamp.getTime())
                .setEndTime(s.getEndTimeUnixMill())
                .setSuccess(s.isSuccess())
                .setErrorMsg(s.StatusMessage)
                .setStatusCode(s.getStatusCode())
                .setType(spanTypeCls)
                .setLayer(layer)
                .addAllTag(tags)
                .setPeer(peer);
        if (s.HasEvent()) {
            String eventLevel = "info";
            String eventMsg = "";
            String eventName = "";
            String stacktrace = "";
            if (s.hasExceptionEvents()) {
                eventLevel = "error";
                TraceException traceException = s.Exceptions.get(0);
                eventMsg = traceException.Message;
                eventName = "exception";
                stacktrace = traceException.StackTrace;
            } else {
                TraceEvent traceEvent = s.Events.get(0);
                eventMsg = traceEvent.Name;
                eventName = traceEvent.Name;
            }
            TraceMessage.Event event = TraceMessage.Event.newBuilder()
                    .setLevel(eventLevel)
                    .setMessage(eventMsg)
                    .setName(eventName)
                    .setStack(stacktrace)
                    .build();
            entrySpanBuilder.setEvent(event);
        }


        TraceMessage.SpanMessage.Builder spanMessageBuilder = TraceMessage.SpanMessage.newBuilder()
                .setMsgType(TraceMessage.SpanMessage.MsgType.apm)
                .setApmSpan(entrySpanBuilder)
                .setSecZoneId("40003102");
        TraceMessage.SpanMessage spanMessage = spanMessageBuilder.build();
        return spanMessage.toByteArray();
    }
}
