package com.cecf.observe;

import com.google.protobuf.ByteString;
import io.opentelemetry.proto.collector.trace.v1.ExportTraceServiceRequest;
import io.opentelemetry.proto.common.v1.InstrumentationScope;
import io.opentelemetry.proto.common.v1.KeyValue;
import io.opentelemetry.proto.resource.v1.Resource;
import io.opentelemetry.proto.trace.v1.*;
import org.apache.flink.api.common.functions.FlatMapFunction;
import org.apache.flink.util.Collector;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.math.BigInteger;
import java.time.Instant;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.cecf.observe.OtelUtils.SpanKindClient;
import static com.cecf.observe.OtelUtils.StatusError;

public class PojoSpanFlatMapFunc implements FlatMapFunction<ExportTraceServiceRequest, PojoSpan> {
    Logger logger = LoggerFactory.getLogger(FlatMapFunction.class);
    Pattern pattern = Pattern.compile("-DAPP_NAME=(\\S+)");

    String getRPCType(int spanKind, KVList attributes) {
        // 2: SERVER, 3: CLIENT, 1: INTERNAL, 4: PRODUCER, 5: CONSUMER
        // http check http.status_code . io.opentelemetry.apache-httpclient-4.0
        // rpc check rpc.system. io.opentelemetry.grpc-1.6
        if (spanKind == 1) {
            return "local";
        }
        if (spanKind == 5) {
            return "consumer";
        }
        if (spanKind == 4) {
            return "producer";
        }
        if (spanKind == 2) {
            // server
            return "server";
        }
        // client: rpc, http
        if (!attributes.getString("http.url").isEmpty()) {
            return "http";
        }
        String rpcSys = attributes.getString("rpc.system");
        if (rpcSys.equals("apache_dubbo")) {
            return "dubbo";
        } else if (!rpcSys.isEmpty()) {
            return rpcSys;
        }
        String dbSys = attributes.getString("db.system");
        if (dbSys.equals("mysql")) {
            return "mysql";
        } else if (!dbSys.isEmpty()) {
            return dbSys;
        }
        return "client";
    }

    String getRPCName(int spanKind, String RPCType, String spanName, KVList attributes) {
        if (spanKind == 3) {
            if (RPCType.equals("dubbo")) {
                return attributes.getString("rpc.method");
            }
            if (RPCType.equals("http")) {
                return attributes.getString("http.target");
            }
        }
        return spanName;
    }

    long getRPCResult(int spanKind, String RPCType, KVList attributes) {
        if (spanKind == 3) {
            if (RPCType.equals("http")) {
                return attributes.getLong("http.status_code");
            }
        }
        return 0;
    }

    @Nullable
    String byteStringToHex(ByteString bs) {
        if (bs == null || bs.isEmpty()) {
            return null;
        }
        String ret = new BigInteger(1, bs.toByteArray()).toString(16);
        return (ret.equals("0") || ret.trim().equals("")) ? null : ret;
    }

    PojoSpan parseSpan(Span span, String serviceName, String scopeName, String serviceIP, String scopeVersion,
                       String containerID, Map<String, String> resourceAttrs) {
        PojoSpan clickhouseSpan = new PojoSpan();
        // appName
        String commandLine = resourceAttrs.get("process.command_line");
        if (commandLine != null && !commandLine.isEmpty()) {
            Matcher matcher = pattern.matcher(commandLine);
            if (matcher.find()) {
                String appAlias = matcher.group(1);
                if (appAlias != null && !appAlias.isEmpty()) {
                    clickhouseSpan.AppName = appAlias;
                }
            }
        }
        if (!clickhouseSpan.hasAppName()) {
            String serviceNamespace = resourceAttrs.get("service.namespace");
            if (serviceNamespace != null && !serviceNamespace.isEmpty() && !serviceNamespace.equals("UNSET")) {
                clickhouseSpan.AppName = serviceNamespace;
            } else {
                String appNameInResource = resourceAttrs.get("app.name");
                if (appNameInResource != null && !appNameInResource.isEmpty() && !appNameInResource.equals("UNSET")) {
                    clickhouseSpan.AppName = appNameInResource;
                }
            }
        }
        // end appName
        clickhouseSpan.TraceID = byteStringToHex(span.getTraceId());
        clickhouseSpan.ResourceAttributes = resourceAttrs;
        clickhouseSpan.SpanID = byteStringToHex(span.getSpanId());
        clickhouseSpan.ParentSpanID = byteStringToHex(span.getParentSpanId());
        clickhouseSpan.SpanKind = span.getKindValue();
        clickhouseSpan.ServiceName = serviceName;
        clickhouseSpan.SpanName = span.getName();
        clickhouseSpan.ScopeName = scopeName;
        clickhouseSpan.ScopeVersion = scopeVersion;
        clickhouseSpan.ContainerID = containerID;
        clickhouseSpan.SrcIP = serviceIP;
        clickhouseSpan.SrcPort = 0;
        clickhouseSpan.TraceState = "";
        clickhouseSpan.RPCRequest = new HashMap<>();

        int statusCode = 0;
        String statusMessage = "";
        if (span.hasStatus()) {
            Status status = span.getStatus();
            statusCode = status.getCodeValue();
            statusMessage = status.getMessage();
        }


        long startTimeNano = span.getStartTimeUnixNano();
        long endTimeNano = span.getEndTimeUnixNano();
        clickhouseSpan.Duration = endTimeNano - startTimeNano;
        clickhouseSpan.Timestamp = Date.from(Instant.ofEpochMilli((long) (startTimeNano / 1e6)));


        List<KeyValue> spanAttributes = span.getAttributesList();
        KVList spanAttributesMap = new KVList(spanAttributes);
        clickhouseSpan.TargetIP = spanAttributesMap.getString("net.peer.name");
        clickhouseSpan.TargetPort = spanAttributesMap.getLong("net.peer.port");
        clickhouseSpan.FuncName = spanAttributesMap.getString("code.function");
        clickhouseSpan.FuncNamespace = spanAttributesMap.getString("code.namespace");
        clickhouseSpan.SpanAttributes = spanAttributesMap.toStringMap();

        if (statusCode == StatusError && statusMessage.isEmpty() && clickhouseSpan.SpanKind == SpanKindClient) {
            // fill status message with http code
            statusMessage = "status code:" + spanAttributesMap.getLong("http.status_code");
        }

        // http check http.status_code . io.opentelemetry.apache-httpclient-4.0
        // rpc check rpc.system. io.opentelemetry.grpc-1.6
        clickhouseSpan.RPCType = this.getRPCType(clickhouseSpan.SpanKind, spanAttributesMap);
        clickhouseSpan.RPCName = this.getRPCName(clickhouseSpan.SpanKind, clickhouseSpan.RPCType,
                clickhouseSpan.SpanName, spanAttributesMap);
        clickhouseSpan.RPCResult = this.getRPCResult(clickhouseSpan.SpanKind, clickhouseSpan.RPCType,
                spanAttributesMap);
        clickhouseSpan.FuncLineNO = spanAttributesMap.getLong("code.lineno");
        clickhouseSpan.DBStatement = spanAttributesMap.getString("db.statement");
        clickhouseSpan.DBConnectionString = spanAttributesMap.getString("db.connection_string");
        for (Span.Event event : span.getEventsList()) {
            long evTs = event.getTimeUnixNano();
            String evName = event.getName();
            if (evName.equals("exception")) {
                KVList exKvs = new KVList(event.getAttributesList());
                String exceptionMessage = exKvs.getString("exception.message");
                String exceptionStack = exKvs.getString("exception.stacktrace");
                String exceptionType = exKvs.getString("exception.type");
                clickhouseSpan.Exceptions.add(new TraceException(exceptionType, exceptionMessage, exceptionStack));
                continue;
            }
            if (evTs > 0 && !evName.isEmpty()) {
                clickhouseSpan.Events.add(new TraceEvent(evTs, evName));
            }
        }

        clickhouseSpan.FuncResult = 0;
        clickhouseSpan.StatusCode = statusCode;
        clickhouseSpan.StatusMessage = statusMessage;
        return clickhouseSpan;
    }

    @Override
    public void flatMap(ExportTraceServiceRequest exportTraceServiceRequest, Collector<PojoSpan> collector) {
        for (ResourceSpans resourceSpans : exportTraceServiceRequest.getResourceSpansList()) {
            Resource resourceAttrs = resourceSpans.getResource();
            KVList resourceKV = new KVList(resourceAttrs.getAttributesList());
            String serviceName = resourceKV.getString("service.name");
            String containerID = resourceKV.getString("container.id");
            String serviceIP = resourceKV.getString("pod.ip");
            for (ScopeSpans scopeSpans : resourceSpans.getScopeSpansList()) {
                InstrumentationScope scope = scopeSpans.getScope();
                KVList scopeKV = new KVList(scope.getAttributesList());
                String scopeName = scopeKV.getString("name");
                String scopeVersion = scopeKV.getString("version", "");
                for (Span span : scopeSpans.getSpansList()) {
                    PojoSpan clickhouseSpan = this.parseSpan(span, serviceName, scopeName, serviceIP,
                            scopeVersion, containerID, resourceKV.toStringMap());
                    collector.collect(clickhouseSpan);
                }
            }
        }
    }
}
