package com.cecf.observe;

import io.opentelemetry.proto.collector.trace.v1.ExportTraceServiceRequest;
import io.opentelemetry.proto.trace.v1.TracesData;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.configuration.RestartStrategyOptions;
import org.apache.flink.connector.kafka.sink.KafkaRecordSerializationSchema;
import org.apache.flink.connector.kafka.sink.KafkaSink;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.ivi.opensource.flinkclickhousesink.ClickHouseSink;
import ru.ivi.opensource.flinkclickhousesink.model.ClickHouseClusterSettings;
import ru.ivi.opensource.flinkclickhousesink.model.ClickHouseSinkConst;

import java.time.Duration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class TraceStream {

    private static String getConfig(String key, String de) {
        String value = System.getProperty(key);
        if (value == null || value.isEmpty()) {
            return de;
        }
        return value;
    }

    private static String[] getConfigArray(String key, String del) {
        if (del == null) {
            del = ",";
        }
        String c = getConfig(key, "");
        if (c == null || c.equals("")) {
            return null;
        }
        return c.split(del);
    }

    private static List<String> getKafkaTopics() {
        String kafkaConf = getConfig("KAFKA_TOPIC", "otel");
        String[] topics = kafkaConf.split(",");
        return Stream.of(topics).map(String::trim).collect(Collectors.toList());
    }

    public static final Logger logger = LoggerFactory.getLogger(TraceStream.class);

    static KafkaSource<String> buildKafkaSource() {
        String brokers = getConfig("KAFKA_BROKERS", "kafka-headless.observe.svc.cluster.local:9092");
        return KafkaSource.<String>builder()
                .setBootstrapServers(brokers)
                .setTopics(getKafkaTopics())
                .setGroupId(getConfig("KAFKA_GROUP", "ob-agent-stream"))
                .setStartingOffsets(OffsetsInitializer.latest())
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }

    public static void main(String[] args) throws Exception {
        Configuration envConfig = new Configuration();
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY, "fixed-delay");
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY_FIXED_DELAY_ATTEMPTS, Integer.MAX_VALUE); // number of restart attempts
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY_FIXED_DELAY_DELAY, Duration.ofSeconds(10)); // delay
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment(envConfig);

        Map<String, String> globalParameters = new HashMap<>();
        String chHosts = getConfig("CH_HOST", "http://clickhouse.observe.svc.cluster.local:8123");
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_HOSTS, chHosts);
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_USER, getConfig("CH_USER", "default"));
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_PASSWORD, getConfig("CH_PWD", "cecf@cestong.com"));

        globalParameters.put(ClickHouseSinkConst.TIMEOUT_SEC, getConfig("CH_TIMEOUT", "10"));
        globalParameters.put(ClickHouseSinkConst.FAILED_RECORDS_PATH, "./failed_record");
        globalParameters.put(ClickHouseSinkConst.NUM_WRITERS, getConfig("CH_WRITER_NUM", "1"));
        globalParameters.put(ClickHouseSinkConst.NUM_RETRIES, getConfig("CH_RETRY_NUM", "3"));
        globalParameters.put(ClickHouseSinkConst.QUEUE_MAX_CAPACITY, getConfig("CH_WRITE_QUEUE_SIZE", "10"));
        globalParameters.put(ClickHouseSinkConst.IGNORING_CLICKHOUSE_SENDING_EXCEPTION_ENABLED, "true");

        String[] excludeServiceList = getConfigArray("EXCLUDE_SERVICE", ",");

        ParameterTool parameters = ParameterTool.fromMap(globalParameters);
        env.getConfig().setGlobalJobParameters(parameters);
        KafkaSource<String> source = buildKafkaSource();

        Properties props = new Properties();
        props.put(ClickHouseSinkConst.TARGET_TABLE_NAME, getConfig("CH_TABLE_NAME", "otel.traces_local"));
        props.put(ClickHouseSinkConst.MAX_BUFFER_SIZE, getConfig("CH_WRITE_BUF_SIZE", "1000000000"));

        PojoSpanSpanCSVConverter converter = new PojoSpanSpanCSVConverter();


        String mysqlDSN = getConfig("MYSQL_DSN", "jdbc:mysql://m1.cestong.com.cn:31511/observe");
        String mysqlUserName = getConfig("MYSQL_USER", "root");
        String mysqlPwd = getConfig("MYSQL_PWD", "pg3mfWRtYonekZWB");

        final OutputTag<PojoSpan> i6000OutputTag = new OutputTag<>("i6000-stream") {
        };

        // todo: 支持oltp proto数据源，合并
        // todo: 使用oltp json直接解析
        SingleOutputStreamOperator<PojoSpan> mainStreamOperator = env.fromSource(source, WatermarkStrategy.noWatermarks(), "KafkaSource")
                .flatMap(new SpringBootSpanExtractMapFunc())
                .name("ResourceSpansFlatToSpan")
                .map(new FillAppNameMapFunc(mysqlDSN, mysqlUserName, mysqlPwd))
                .name("AttachAppName")
                .filter(new ServiceNameFilter(excludeServiceList))
                .name("FilterByServiceName")
                .process(new ProcessFunction<PojoSpan, PojoSpan>() {
                    @Override
                    public void processElement(PojoSpan pojoSpan, ProcessFunction<PojoSpan, PojoSpan>.Context context, Collector<PojoSpan> collector) throws Exception {
                        collector.collect(pojoSpan);
                        context.output(i6000OutputTag, pojoSpan);
                    }
                })
                .name("SideOutput");

        // i6000 stream
        String[] includeI6000AppNames = getConfigArray("i6000.include_app_names", ",");
        String[] excludeI6000ServiceNames = getConfigArray("i6000.exclude_service_names", ",");
        KafkaSink<byte[]> i6000KafkaSink = getSpringBootSpanKafkaSink();
        mainStreamOperator.getSideOutput(i6000OutputTag)
                .filter(new AppNameFilter(includeI6000AppNames))
                .name("filterByAppName")
                .filter(new ServiceNameFilter(excludeI6000ServiceNames))
                .name("FilterI6000AppNames")
//                .filter(new SpanKindFilter(SpanKindClient, SpanKindServer, SpanKindInternal))
//                .name("FilterSpanKind")
                .map(new PojoSpan2TracesMessage())
                .name("PojoSpan2Bytes")
                .sinkTo(i6000KafkaSink)
                .name("i6000-kafka");

        mainStreamOperator
                .addSink(new ClickHouseSink<>(props, converter))
                .name("ClickhouseSink");

        logger.info("配置完成，运行flink");
        env.execute(getConfig("FLINK_JOB_NAME", "ob-agent-stream"));
    }

    private static KafkaSink<byte[]> getSpringBootSpanKafkaSink() {
        String i6000KafkaBootstrapServer = getConfig("i6000.kafka.bootstrap-server", "kafka.observer.svc.cluster.local:9092");
        return KafkaSink.<byte[]>builder()
                .setBootstrapServers(i6000KafkaBootstrapServer)
                .setRecordSerializer(
                        KafkaRecordSerializationSchema.builder()
                                .setTopic(getConfig("i6000.kafka.topic", "i6000_trace"))
                                .setValueSerializationSchema(new I6000TraceMessageSerializer()).build()
                )
                .build();
    }


}
